use std::fmt::Display;
use std::net::SocketAddr;

use anyhow::Result;
use async_trait::async_trait;
use libpt::log::{debug, info, warn};
use rand::distributions::{Distribution, Standard};
use rand::{random, Rng};
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};
use tokio::net::{TcpListener, TcpStream};
use tokio::time::Instant;

use super::Challenge;
use crate::config::Config;
use crate::has_won;
use crate::vault::VaultRef;

const NEEDED_CORRECT: usize = 16;

#[derive(Copy, Clone)]
enum Operation {
    Add,
    Sub,
    Div,
    Mul,
}

impl Operation {
    pub(crate) fn calc(self, a: i32, b: i32) -> i32 {
        match self {
            Self::Add => a.overflowing_add(b).0,
            Self::Sub => a.overflowing_sub(b).0,
            Self::Div => a.overflowing_div(b).0,
            Self::Mul => a.overflowing_mul(b).0,
        }
    }
}

impl Display for Operation {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Self::Add => "+",
                Self::Sub => "-",
                Self::Mul => "*",
                Self::Div => "/",
            }
        )
    }
}

impl Distribution<Operation> for Standard {
    fn sample<R: rand::Rng + ?Sized>(&self, rng: &mut R) -> Operation {
        match rng.gen_range(0..3) {
            0 => Operation::Add,
            1 => Operation::Sub,
            2 => Operation::Mul,
            3 => Operation::Div,
            _ => unreachable!(),
        }
    }
}

pub struct C3 {
    config: Config,
    vault: VaultRef,
}

impl C3 {
    fn mk_question() -> (String, String) {
        let mut rng = rand::thread_rng();
        let a: i32 = rng.gen_range(0i32..i16::MAX as i32);
        let b: i32 = rng.gen_range(0i32..i16::MAX as i32);
        let op: Operation = random();
        let res = op.calc(a, b);

        let question = format!("What is {a} {op} {b}?");
        let solution = format!("{}", res);
        (question, solution)
    }

    async fn handler(peer: (TcpStream, SocketAddr), vault: &VaultRef) -> Result<()> {
        let mut stream = peer.0;
        let addr = peer.1;
        let mut correct: usize = 0;

        let mut buf: Vec<u8> = Vec::new();

        let (stream_read, mut stream_write) = stream.split();
        let mut rdbuf = BufReader::new(stream_read);

        let mut last_reset = Instant::now();

        loop {
            if correct >= NEEDED_CORRECT {
                Self::win(vault, &mut stream, &addr).await?;
                break;
            }
            if last_reset.elapsed().as_millis() > NEEDED_CORRECT as u128 * 200 {
                correct = 0;
                debug!("{addr} was too slow, resetting");
                stream_write
                    .write_all("You're too slow!\n".to_string().as_bytes())
                    .await?;
                last_reset = Instant::now();
                continue;
            }
            let (question, solution) = Self::mk_question();
            debug!("sending question to {addr}: '{question}'");
            stream_write.write_all((question + "\n").as_bytes()).await?;
            rdbuf.read_until(0x0A, &mut buf).await?; // read until a newline comes
            let answer = String::from_utf8_lossy(&buf).trim().to_string();
            if answer == solution {
                correct += 1;
                debug!("{addr} got one correct");
                stream_write
                    .write_all(format!("correct! {correct} / {NEEDED_CORRECT}\n").as_bytes())
                    .await?;
            } else {
                debug!("{addr} screwed up, {answer} != {solution}");
                stream_write
                    .write_all("wrong!\n".to_string().as_bytes())
                    .await?;
                correct = 0;
                last_reset = Instant::now();
            }
            buf.clear()
        }

        Ok(())
    }

    async fn win(
        vault: &VaultRef,
        stream: &mut TcpStream,
        addr: &std::net::SocketAddr,
    ) -> Result<()> {
        has_won(addr);
        if let Err(e) = stream
            .write_all(("You win: ".to_owned() + vault.secret()).as_bytes())
            .await
        {
            warn!("could not write to peer {addr}: {e}");
            return Err(e.into());
        };
        if let Err(e) = stream.shutdown().await {
            warn!("could end connection to peer {addr}: {e}");
            return Err(e.into());
        };
        Ok(())
    }
}

#[async_trait]
impl Challenge for C3 {
    fn new(config: Config, vault: VaultRef) -> Self {
        info!("Solution: {}", Self::solution());
        Self { config, vault }
    }

    fn hints() -> Vec<String> {
        vec![
            "TCP connect to 1337 and answer the questions".to_string(),
            "The questions keep chaning".to_string(),
            "You should try to solve the questions in an automated way".to_string(),
        ]
    }

    fn solution() -> String {
        String::from("Connect by TCP, send 1337 as bytes (not text).")
    }

    async fn serve(self) -> anyhow::Result<()> {
        info!("serving challenge 3");
        let listener = TcpListener::bind(self.config.addr).await?;

        loop {
            let vault = self.vault.clone();
            let peer = match listener.accept().await {
                Ok(s) => s,
                Err(err) => {
                    warn!("could not accept tcp stream: {err:?}");
                    continue;
                }
            };
            info!("new peer: {}", peer.1);
            tokio::spawn(async move {
                if let Err(e) = Self::handler(peer, &vault).await {
                    warn!("error while handling peer: {e}");
                }
            });
        }
    }
}
